// DlgNetlist.cpp : implementation file
//

#include "stdafx.h"
#include "FreePcb.h"
#include "DlgNetlist.h"
#include "DlgEditNet.h"
#include "DlgSetTraceWidths.h"
#include "DlgNetCombine.h"
#include "DlgChooseNetName.h"

// columns for list
enum {
	COL_NAME = 0,
	COL_PINS,
	COL_WIDTH,
	COL_VIA_W,
	COL_HOLE_W
};

// sort types
enum {
	SORT_UP_NAME = 0,
	SORT_DOWN_NAME,
	SORT_UP_PINS,
	SORT_DOWN_PINS,
	SORT_UP_WIDTH,
	SORT_DOWN_WIDTH,
	SORT_UP_VIA_W,
	SORT_DOWN_VIA_W,
	SORT_UP_HOLE_W,
	SORT_DOWN_HOLE_W
};

// global so that it is available to Compare() for sorting list control items
netlist_info nl_combine;

// global callback function for sorting
// lp1, lp2 are indexes to global arrays above
//		
int CALLBACK CompareNetlistCombine( LPARAM lp1, LPARAM lp2, LPARAM type )
{
	int ret = 0;
	switch( type )
	{
		case SORT_UP_NAME:
		case SORT_DOWN_NAME:
			ret = (_tcscmp( ::nl_combine[lp1].name, ::nl_combine[lp2].name ));
			break;

		case SORT_UP_WIDTH:
		case SORT_DOWN_WIDTH:
			if( ::nl_combine[lp1].w > ::nl_combine[lp2].w )
				ret = 1;
			else if( ::nl_combine[lp1].w < ::nl_combine[lp2].w )
				ret = -1;
			break;

		case SORT_UP_VIA_W:
		case SORT_DOWN_VIA_W:
			if( ::nl_combine[lp1].v_w > ::nl_combine[lp2].v_w )
				ret = 1;
			else if( ::nl_combine[lp1].v_w < ::nl_combine[lp2].v_w )
				ret = -1;
			break;

		case SORT_UP_HOLE_W:
		case SORT_DOWN_HOLE_W:
			if( ::nl_combine[lp1].v_h_w > ::nl_combine[lp2].v_h_w )
				ret = 1;
			else if( ::nl_combine[lp1].v_h_w < ::nl_combine[lp2].v_h_w )
				ret = -1;
			break;

		case SORT_UP_PINS:
		case SORT_DOWN_PINS:
			if( ::nl_combine[lp1].ref_des.GetSize() > ::nl_combine[lp2].ref_des.GetSize() )
				ret = 1;
			else if( ::nl_combine[lp1].ref_des.GetSize() < ::nl_combine[lp2].ref_des.GetSize() )
				ret = -1;
			break;
	}
	switch( type )
	{
		case SORT_DOWN_NAME:
		case SORT_DOWN_WIDTH:
		case SORT_DOWN_VIA_W:
		case SORT_DOWN_HOLE_W:
		case SORT_DOWN_PINS:
			ret = -ret;
			break;
	}

	return ret;
}

// CDlgNetCombine dialog

IMPLEMENT_DYNAMIC(CDlgNetCombine, CDialog)

/*=============================================================================
					CDlgNetCombine NX̍\z
=============================================================================*/

CDlgNetCombine::CDlgNetCombine(CWnd* pParent /*=NULL*/)
	: CDialog(CDlgNetCombine::IDD, pParent)
{
	//{{AFX_DATA_INIT(CDlgNetCombine)
	//}}AFX_DATA_INIT

	m_w = 0;
	m_v_w = 0;
	m_v_h_w = 0;
	m_nlist = 0;
}

/*=============================================================================
					CDlgNetCombine NX̏
=============================================================================*/

CDlgNetCombine::~CDlgNetCombine()
{
}

/*=============================================================================
			CDlgNetCombine _CAOf[^̌ƗL`FbN
=============================================================================*/

void	CDlgNetCombine::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CDlgNetCombine)
	//}}AFX_DATA_MAP

	DDX_Control(pDX, IDC_LIST_NET, m_wndListCtrl);
	if(pDX->m_bSaveAndValidate)				/* Rg[ϐ֒l擾	*/
	{
	}
	else									/* ϐRg[	*/
	{
	}
}

/*=============================================================================
				CDlgNetCombine NX̃bZ[W}bv
=============================================================================*/

BEGIN_MESSAGE_MAP(CDlgNetCombine, CDialog)
	//{{AFX_MSG_MAP(CDlgNetCombine)
	//}}AFX_MSG_MAP
	ON_NOTIFY(LVN_COLUMNCLICK, IDC_LIST_NET, OnLvnColumnclickListNet)
	ON_BN_CLICKED(IDOK, OnBnClickedOk)
	ON_BN_CLICKED(IDCANCEL, OnBnClickedCancel)
END_MESSAGE_MAP()

/*=============================================================================

=============================================================================*/

void	CDlgNetCombine::Initialize(CNetList * nlist				/**< ToDo:Add Comment for Doxygen			*/
								  ,CPartList * plist			/**< ToDo:Add Comment for Doxygen			*/ )
{
	m_nlist = nlist;
	m_plist = plist;
}

// CDlgNetCombine message handlers

/*=============================================================================
					_CAȌ
=============================================================================*/

BOOL	CDlgNetCombine::OnInitDialog()
{
	CDialog::OnInitDialog();

	// make copy of netlist data so that it can be edited but user can still cancel
	// use global netlist_info so it can be sorted in the list control
	m_nl = &::nl_combine;
	m_nlist->ExportNetListInfo( m_nl );

	// initialize netlist control
	m_item_selected = -1;
	m_sort_type = 0;
	m_visible_state = 1;
	DrawListCtrl();
	return TRUE;  
}

// draw listview control and sort according to m_sort_type

/*=============================================================================

=============================================================================*/

void	CDlgNetCombine::DrawListCtrl()
{
	int nItem;
	CString str;
	DWORD old_style = m_wndListCtrl.GetExtendedStyle(); 
	m_wndListCtrl.SetExtendedStyle( LVS_EX_FULLROWSELECT | LVS_EX_FLATSB | old_style );
	m_wndListCtrl.DeleteAllItems();
	m_wndListCtrl.InsertColumn(COL_NAME	, _T("Name")	, LVCFMT_LEFT, 140 );
	m_wndListCtrl.InsertColumn(COL_PINS	, _T("Pins")	, LVCFMT_LEFT, 40 );
	m_wndListCtrl.InsertColumn(COL_WIDTH	, _T("Width")	, LVCFMT_LEFT, 40 );
	m_wndListCtrl.InsertColumn(COL_VIA_W	, _T("Via W")	, LVCFMT_LEFT, 40 );   
	m_wndListCtrl.InsertColumn(COL_HOLE_W	, _T("Hole")	, LVCFMT_LEFT, 40 );
	int iItem = 0;
	for( int i=0; i<::nl_combine.GetSize(); i++ )
	{
			nItem = m_wndListCtrl.InsertItem( iItem, _T(""));
			m_wndListCtrl.SetItemData( iItem, (LPARAM)i );
			m_wndListCtrl.SetItem( iItem, COL_NAME, LVIF_TEXT, ::nl_combine[i].name, 0, 0, 0, 0 );
			str.Format(_T("%d"), ::nl_combine[i].ref_des.GetSize() );
			m_wndListCtrl.SetItem( iItem, COL_PINS, LVIF_TEXT, str, 0, 0, 0, 0 );
			str.Format(_T("%d"), ::nl_combine[i].w/NM_PER_MIL );
			m_wndListCtrl.SetItem( iItem, COL_WIDTH, LVIF_TEXT, str, 0, 0, 0, 0 );
			str.Format(_T("%d"), ::nl_combine[i].v_w/NM_PER_MIL );
			m_wndListCtrl.SetItem( iItem, COL_VIA_W, LVIF_TEXT, str, 0, 0, 0, 0 );
			str.Format(_T("%d"), ::nl_combine[i].v_h_w/NM_PER_MIL );
			m_wndListCtrl.SetItem( iItem, COL_HOLE_W, LVIF_TEXT, str, 0, 0, 0, 0 );
			ListView_SetCheckState( m_wndListCtrl, nItem, ::nl_combine[i].visible );
	}
	m_wndListCtrl.SortItems( ::CompareNetlistCombine, m_sort_type );
}

/*=============================================================================

=============================================================================*/

void	CDlgNetCombine::OnLvnColumnclickListNet(NMHDR *pNMHDR, LRESULT *pResult)
{
	LPNMLISTVIEW pNMLV = reinterpret_cast<LPNMLISTVIEW>(pNMHDR);
	int column = pNMLV->iSubItem;
	if( column == COL_NAME )
	{
		if( m_sort_type == SORT_UP_NAME )
			m_sort_type = SORT_DOWN_NAME;
		else
			m_sort_type = SORT_UP_NAME;
		m_wndListCtrl.SortItems( ::CompareNetlistCombine, m_sort_type );
	}
	else if( column == COL_WIDTH )
	{
		if( m_sort_type == SORT_UP_WIDTH )
			m_sort_type = SORT_DOWN_WIDTH;
		else
			m_sort_type = SORT_UP_WIDTH;
		m_wndListCtrl.SortItems( ::CompareNetlistCombine, m_sort_type );
	}
	else if( column == COL_VIA_W )
	{
		if( m_sort_type == SORT_UP_VIA_W )
			m_sort_type = SORT_DOWN_VIA_W;
		else
			m_sort_type = SORT_UP_VIA_W;
		m_wndListCtrl.SortItems( ::CompareNetlistCombine, m_sort_type );
	}
	else if( column == COL_HOLE_W )
	{
		if( m_sort_type == SORT_UP_HOLE_W )
			m_sort_type = SORT_DOWN_HOLE_W;
		else
			m_sort_type = SORT_UP_HOLE_W;
		m_wndListCtrl.SortItems( ::CompareNetlistCombine, m_sort_type );
	}
	else if( column == COL_PINS )
	{
		if( m_sort_type == SORT_UP_PINS )
			m_sort_type = SORT_DOWN_PINS;
		else
			m_sort_type = SORT_UP_PINS;
		m_wndListCtrl.SortItems( ::CompareNetlistCombine, m_sort_type );
	}
	*pResult = 0;
}

/*=============================================================================

=============================================================================*/

void	CDlgNetCombine::OnBnClickedOk()
{
	TCHAR	ch[41];

	int n = m_wndListCtrl.GetSelectedCount();
	if( n == 0 )
	{
		AfxMessageBox(_T("No nets selected"));
		return;
	}
	m_names.SetSize(n);
	int i = 0;
	POSITION pos = m_wndListCtrl.GetFirstSelectedItemPosition();
	while( pos )
	{
		int nItem = m_wndListCtrl.GetNextSelectedItem(pos);
		m_wndListCtrl.GetItemText( nItem, 0, ch, 40 );
		m_names[i++] = ch;
	}
	CDlgChooseNetName dlg;
	dlg.Initialize( &m_names );
	int ret = dlg.DoModal();
	if( ret == IDOK )
	{
		m_new_name = dlg.m_sel_str;
		OnOK();
	}
	else
	{
		return;
	}
}

/*=============================================================================

=============================================================================*/

void	CDlgNetCombine::OnBnClickedCancel()
{
	// TODO: Add your control notification handler code here
	OnCancel();
}

