// classes to manage footprint libraries
//
#pragma once
#include "DlgLog.h"

// this class represents a single footprint in a file

/*=============================================================================
					NX̐錾
=============================================================================*/

class CFootLibEntry
{
public:
	long m_offset;		// position in file
	CString m_name;		// footprint name
	long m_next_offset;	// offset to next heading or footprint in file
};

/*=============================================================================
					NX̐錾
=============================================================================*/

// this class represents a footprint library file
class CFootLib
{
//--------------------------------------
// Construction/Destruction - y\z/Łz
//--------------------------------------
public:
	CFootLib(){ m_indexed = m_expanded = FALSE; };
	CString m_file_name;	// filename
	CString m_full_path;	// path
	BOOL m_expanded;		// save expanded state in tree control
	BOOL m_indexed;			// TRUE if library has been indexed
	CArray<CFootLibEntry> m_foot;	// footprints 
};

/*=============================================================================

=============================================================================*/

// this class represents a footprint library folder,
// containing multiple footprint library files
class CFootLibFolder
{
public:
//--------------------------------------
// Construction/Destruction - y\z/Łz
//--------------------------------------
	~CFootLibFolder(){ Clear(); };
	void Clear();
	void IndexAllLibs( CString * full_path, CDlgLog * dlg_log );
	int GetNumLibs() const;
	void IndexLib( CString * file_name, CDlgLog * dlog = NULL );
	LPCTSTR		GetFullPath() const;
	int SearchFileName( CString * fn );
	BOOL GetFootprintInfo( CString * name, CString * prefer_lib_file,
		int * ilib, int * ifootprint, CString * file_name, 
		int * offset, int * next_offset = NULL ) const;
	LPCTSTR		GetLibraryFileName( int ilib ) const;
	LPCTSTR		GetLibraryFullPath( int ilib ) const;
	int GetNumFootprints( int ilib ) const;
	LPCTSTR		GetFootprintName( int ilib, int ifoot ) const;
	int GetFootprintOffset( int ilib, int ifoot ) const;
	void SetExpanded( int ilib, BOOL state ){ m_footlib[ilib].m_expanded = state; };
	BOOL GetExpanded( int ilib ) const	{ return m_footlib[ilib].m_expanded; };
private:
	CString m_full_path_to_folder; 
	CArray<CFootLib> m_footlib;	// array of library files
	CMapStringToPtr m_lib_map;	// map of all footprint names
};

/*=============================================================================
					NX̐錾
=============================================================================*/

// this class keeps track of multiple footprint library folders
// basically just maps CFootLibFolders to path names
class CFootLibFolderMap
{
//--------------------------------------
// Construction/Destruction - y\z/Łz
//--------------------------------------
public:
	CFootLibFolderMap();
	~CFootLibFolderMap();
	void AddFolder(LPCTSTR lpszFullPath,CFootLibFolder * folder );
	CFootLibFolder * GetFolder(LPCTSTR lpszFullPath);
	BOOL FolderIndexed( CString * full_path );
	void SetDefaultFolder(LPCTSTR lpszDefaultFullPath);
	void SetLastFolder(LPCTSTR lpszLastFullPath);
	const CString	&GetDefaultFolder() const;
	const CString	&GetLastFolder() const;
private:
	CMapStringToPtr		m_folder_map;			// map of folders
	CString				m_default_folder;		// path to default folder
	CString				m_last_folder;			// path to last folder used
};

//{{AFX_INSERT_LOCATION}}
// Microsoft Visual C++ will insert additional declarations immediately before the previous line.
/*-----------------------------------------------------------------------------
						end of FootprintLib.h
-----------------------------------------------------------------------------*/
