// functions to convert Ivex .mod file to FreePCB .lib file
//
#include "stdafx.h"
#include "file_io.h"
#include "Shape.h"
#include "ivex_mod_file.h"
#include "math.h"

// define layer masks
enum {
	LAY_T = 0x1,
	LAY_I = 0x2,
	LAY_B = 0x4,
	LAY_C = 0x8,
	LAY_P = 0x10
};

// structures to hold padstack data from mod file
struct pad_stack_entry {
	CString shape;
	int pad_shape;
	int orientation;
	int x_wid, y_wid;
	int x_off, y_off;
	int layer_mask;
};

struct mod_padstack
{
	CString					name;			/* Pad stack name			*/
	int						hole_diam;		// Drill dia in units = NM	*/
	CArray<pad_stack_entry>	entry;			/*							*/
};
typedef mod_padstack ps;

ps * current_ps;

// FreePCB shape

// this is the main conversion function
//

/*=============================================================================

=============================================================================*/

int	ConvertIvex(LPCTSTR lpszInputFileName		/*=NULL*/	/**< ToDo:									*/
			   ,LPCTSTR lpszOutputFileName		/*=NULL*/	/**< ToDo:									*/
			   ,CEdit	*edit_ctrl							/**< ToDo:									*/ )
//			   ,int (CALLBACK *Callback)(INT nLine,INT nBlock,LPVOID pObject) /*=NULL*/	/**< ɌĂяoR[obN֐̃AhX	*/
//			   ,LPVOID pObject				/*=NULL */	/**< ToDo:										*/ )
{
	CString		strInputFilePathName;
	CString		strOutputFilePathName;
	CStdioFile	objInputFile;
	CStdioFile	objOutputFile;

	if(lpszInputFileName && lstrlen(lpszInputFileName))
		strInputFilePathName=lpszInputFileName;
	else
	{
		CFileDialog wndDlg(TRUE
						,_T(".mod")
						,_T("*.mod")
						,OFN_FILEMUSTEXIST|OFN_ENABLESIZING|OFN_NOCHANGEDIR|OFN_SHOWHELP
						,_T("Ivex library files (*.mod)|*.mod|All Files (*.*)|*.*||"));

		if(wndDlg.DoModal()==IDOK)
			strInputFilePathName=wndDlg.GetPathName();
		else
			return -1;
	}
	if(lpszOutputFileName && lstrlen(lpszOutputFileName))
		strOutputFilePathName=lpszOutputFileName;
	else
	{
		CFileDialog	wndDlg(FALSE																/*									*/
						,_T(".fpl")																/*									*/
						,_T("*.fpl")															/*									*/
						,OFN_CREATEPROMPT														/*									*/
						|OFN_OVERWRITEPROMPT													/*									*/
						|OFN_HIDEREADONLY														/*									*/
						|OFN_ENABLESIZING														/*									*/
						|OFN_NOCHANGEDIR														/*									*/
						|OFN_SHOWHELP															/*									*/
						,_T("Free PCB library files (*.fpl)|*.fpl|All Files (*.*)|*.*||"));		/*									*/
		if(wndDlg.DoModal()==IDOK)
			strOutputFilePathName=wndDlg.GetPathName();
		else
			return -1;
	}
	int err = objInputFile.Open(strInputFilePathName,CFile::modeRead,NULL);
	if( !err )
	{
		// error opening file
		CString	strErrMsg;

		strErrMsg.Format(_T("Unable to open input file %s"),strInputFilePathName);
		AfxMessageBox(strErrMsg,MB_ICONWARNING|MB_OK);
		return -2;
	}
	err = objOutputFile.Open(strOutputFilePathName , CFile::modeWrite|CFile::modeCreate, NULL );
	if( !err )
	{
		// error opening file
		CString strErrMsg;
		strErrMsg.Format(_T("Unable to open output file %s"),strOutputFilePathName);
		AfxMessageBox(strErrMsg,MB_ICONWARNING|MB_OK);
		return -3;
	}
	CString str;

	str.Format(_T("Converting Ivex file %s...\r\n"), strInputFilePathName );
	edit_ctrl->ReplaceSel( str );
	err = ConvertIvex( &objInputFile, &objOutputFile, edit_ctrl);
	edit_ctrl->ReplaceSel(_T("done\r\n"));
	objInputFile.Close();
	objOutputFile.Close();
	return(0);
}

/*=============================================================================

=============================================================================*/

int	ConvertIvex(CStdioFile * mod_file		/**< ToDo:									*/
			   ,CStdioFile * lib_file		/**< ToDo:									*/
			   ,CEdit * edit_ctrl			/**< ToDo:									*/ )
//			   ,int (CALLBACK *Callback)(INT nLine,INT nBlock,LPVOID pObject) /*=NULL*/	/**< ɌĂяoR[obN֐̃AhX	*/
//			   ,LPVOID pObject	/*=NULL */	/**< ToDo:										*/ )
{
	// state-machine
	enum { IDLE,	// idle (i.e. no state)
		PAD_STACK,	// parsing pad stack
		MODULE		// parsing module
	};

	CArray<mtg_hole> m_mtg_hole;	
	CShape			s;
	s.m_padstack.SetSize(0);
	s.m_outline_poly.SetSize(0);
	m_mtg_hole.SetSize(0);
	s.m_author = _T("Ivex");

	CPoint		ptMinPos;							// Replaced integer variable (min_x,min_y) to a CPoint object.
	CPoint		ptMaxPos;							// Replaced integer variable (max_x,max_y) to a CPoint object.

#define pi  3.14159265359	

	double sin45 = sin( pi/4.0 );
	double cos45 = sin45;
	double sin22 = sin( pi/8.0 );
	double cos22 = cos( pi/8.0 );
	double sin67 = sin( 3.0*pi/8.0 );
	double cos67 = cos( 3.0*pi/8.0 );

	// state
	int			state = IDLE;
	int			units = MIL;
	int			mult = NM_PER_MIL;
	CString		instr, keystr, subkey;
	CArray<CString> p;
	mod_padstack * new_ps;
	int			current_x, current_y;
	int			current_angle;

	// map of pad identifiers
	CMapStringToPtr mod_padstack_map;
	CString			strPadStackName;
	CString			strDataSource,strManufacture;

	// start reading mod file
	int ret = mod_file->ReadString( instr );
	instr.Trim();
	while( ret )
	{
		instr.MakeUpper();
		if(instr.Left(9)==_T("// SOURCE"))
		{
			// get source of footprint
			strDataSource=instr.Right(instr.GetLength()-9);
			strDataSource.Remove(_T('\"'));
			strDataSource.Trim();
			if(!strDataSource.IsEmpty())
				s.m_source = strDataSource;
		}
		else if(instr.Left(16)==_T("// DATA SOURCE :"))
		{
			strDataSource=instr.Right(instr.GetLength()-16);
			strDataSource.Remove(_T('\"'));
			strDataSource.Trim();
			if(!strDataSource.IsEmpty())
				s.m_source = strDataSource;
		}
		else if(instr.Left(16)==_T("// MANUFACTURE :"))
		{
			strManufacture=instr.Right(instr.GetLength()-16);
			strManufacture.Remove(_T('\"'));
			strManufacture.Trim();
			if(!strManufacture.IsEmpty())
				s.m_desc = strManufacture;
		}
		int np = ParseKeyString( &instr, &keystr, &p ); 
		if( np )
		{
			// now handle incoming line
//			if( state == IDLE && keystr == _T("UNITS"))
			if(keystr == _T("UNITS"))
			{
				subkey = p[0];
				if( subkey ==_T("MIL"))
				{
					units = MIL;
					mult = NM_PER_MIL;
				}
				else if( subkey==_T("MM"))
				{
					units = MM;
					mult = 1000000;
				}
				else if( subkey == _T("NM"))
				{
					units = NM;
					mult = 1;
				}
				else if( subkey == _T("MICRON"))
				{
					units = MM;
					mult = 1000;
				}
				else
					ASSERT(FALSE);				/* Unknown measure unit parameter! */
			}
			else if( state == IDLE && keystr == _T("PAD") && p[0] == _T("STACK"))
			{
				// parse PAD STACK
				state = PAD_STACK;
				strPadStackName = p[1];
				double hole_diam = my_atof(p[2]);
				if( p[3] != _T("{"))
					ASSERT(FALSE);
				// store it in map
				new_ps = new mod_padstack;
				new_ps->name = strPadStackName;
				new_ps->hole_diam = hole_diam * mult;
				mod_padstack_map.SetAt( strPadStackName, new_ps );
			}
			else if( state == PAD_STACK && ( keystr == _T("OVAL") || keystr == _T("RECT")))
			{
				// parse PAD STACK entry
				int i = new_ps->entry.GetSize();

				new_ps->entry.SetSize( i+1 );
				new_ps->entry[i].shape = keystr;
				new_ps->entry[i].pad_shape = PAD_NONE;
				if( keystr == _T("OVAL"))
					new_ps->entry[i].pad_shape = PAD_OVAL;
				else if( keystr == _T("RECT"))
					new_ps->entry[i].pad_shape = PAD_RECT;
				else
					ASSERT(FALSE);
				new_ps->entry[i].orientation = my_atof(p[0]);
				new_ps->entry[i].x_wid = my_atof(p[1])*mult;				/* Pad size (x)				*/
				new_ps->entry[i].y_wid = my_atof(p[2])*mult;				/* Pad size (y)				*/
				if(new_ps->entry[i].x_wid==new_ps->entry[i].y_wid)
				{
					if(new_ps->entry[i].pad_shape==PAD_RECT)
						new_ps->entry[i].pad_shape=PAD_SQUARE;
					else if(new_ps->entry[i].pad_shape==PAD_OVAL)
						new_ps->entry[i].pad_shape=PAD_ROUND;
				}
				new_ps->entry[i].x_off = my_atof(p[3])*mult;				/* Pad center offset(x)		*/
				new_ps->entry[i].y_off = my_atof(p[4])*mult;				/* Pad center offset(y)		*/
				CString layers = p[5];
				if( np >= 8 )
					layers += p[6];
				if( np >= 9 )
					layers += p[7];
				new_ps->entry[i].layer_mask = 0;
				if( layers.Find('T') != -1 )
					new_ps->entry[i].layer_mask |= LAY_T;
				if( layers.Find('I') != -1 )
					new_ps->entry[i].layer_mask |= LAY_I;
				if( layers.Find('B') != -1 )
					new_ps->entry[i].layer_mask |= LAY_B;
				if( layers.Find('C') != -1 )
					new_ps->entry[i].layer_mask |= LAY_C;
				if( layers.Find('P') != -1 )
					new_ps->entry[i].layer_mask |= LAY_P;
			}
			else if( state == PAD_STACK && keystr == _T("}"))
			{
#ifdef _DEBUG
				int						nLastPasdStack;
				void					*ptr;
				const mod_padstack		*pPadStack;

				TRACE(_T("\r\n*** %s ***"),strPadStackName);
				if(mod_padstack_map.Lookup(strPadStackName,ptr))
				{
					pPadStack=(const mod_padstack *)ptr;
					for(int i=0;i<pPadStack->entry.GetSize();i++)
					{
						TRACE(_T("\r\n< %s > - [%d] %d (%d,%d) (%d,%d) %d")
							,pPadStack->entry[i].shape
							,pPadStack->entry[i].pad_shape
							,pPadStack->entry[i].orientation
							,pPadStack->entry[i].x_wid
							,pPadStack->entry[i].y_wid
							,pPadStack->entry[i].x_off
							,pPadStack->entry[i].y_off
							,pPadStack->entry[i].layer_mask);
					}
				}
				else
				{
					ASSERT(FALSE);
				}
#endif
				state = IDLE;
			}
			else if( state == IDLE && keystr == _T("BEGIN"))
			{
				// start new module (i.e. footprint) definition
				state = MODULE;
				s.m_units = units;
				s.m_padstack.SetSize(0);
				s.m_outline_poly.SetSize(0);
				m_mtg_hole.SetSize(0);
				// default ref text params
				s.m_ref_size = 50 * NM_PER_MIL;
				s.m_ref_xi = 50 * NM_PER_MIL;
				s.m_ref_yi = 50 * NM_PER_MIL;
				s.m_ref_angle = 0;
				s.m_ref_w = 7 * NM_PER_MIL;
				// track min and max dimensions to create selection rectangle
				ptMinPos = CPoint(INT_MAX,INT_MAX);
				ptMaxPos = CPoint(INT_MIN,INT_MIN);
			}
			else if( state == MODULE && keystr == _T("END"))
			{
				// end module def.
				state = IDLE;
				// check for all pins defined
				BOOL error = FALSE;
				for( int ip=0; ip<s.m_padstack.GetSize(); ip++ )
				{
					if( !s.m_padstack[ip].exists )
					{
						error = TRUE;
						break;
					}
				}
				if( !error )
				{
					// convert mounting holes to pads
					int ip = s.m_padstack.GetSize();
					s.m_padstack.SetSize( ip + m_mtg_hole.GetSize() );
					for( int ih=0; ih<m_mtg_hole.GetSize(); ih++ )
					{
						s.m_padstack[ip].exists = 1;
						s.m_padstack[ip].name.Format(_T("MH%d"), ih+1 );
						s.m_padstack[ip].hole_size = m_mtg_hole[ih].diam;
						s.m_padstack[ip].x_rel = m_mtg_hole[ih].x;
						s.m_padstack[ip].y_rel = m_mtg_hole[ih].y;
						s.m_padstack[ip].angle = 0;
						s.m_padstack[ip].top.shape = m_mtg_hole[ih].pad_shape;
						s.m_padstack[ip].top.size_h = m_mtg_hole[ih].pad_diam;
						s.m_padstack[ip].top.size_l = 0;
						s.m_padstack[ip].top.size_r = 0;
						int inner_pad_diam;
						if( m_mtg_hole[ih].pad_shape == PAD_NONE )
							s.m_padstack[ip].inner.shape = PAD_NONE;
						else
						{
							s.m_padstack[ip].inner.shape = PAD_ROUND;
							inner_pad_diam = m_mtg_hole[ih].diam + 20*NM_PER_MIL;
							if( inner_pad_diam < m_mtg_hole[ih].pad_diam )
								s.m_padstack[ip].inner.size_h = inner_pad_diam;
							else
								s.m_padstack[ip].inner.size_h = m_mtg_hole[ih].pad_diam;
							s.m_padstack[ip].inner.size_l = 0;
							s.m_padstack[ip].inner.size_r = 0;
						}
						s.m_padstack[ip].bottom.shape = m_mtg_hole[ih].pad_shape;
						s.m_padstack[ip].bottom.size_h = m_mtg_hole[ih].pad_diam;
						s.m_padstack[ip].bottom.size_l = 0;
						s.m_padstack[ip].bottom.size_r = 0;
						ip++;
					}
					m_mtg_hole.RemoveAll();
					// make sure selection rectangle surrounds pads
					for( int ip=0; ip<s.m_padstack.GetSize(); ip++ )
					{
						int x = s.m_padstack[ip].x_rel;
						int y = s.m_padstack[ip].y_rel;
						int dx = s.m_padstack[ip].top.size_h/2;
						int dy = s.m_padstack[ip].top.size_h/2;
						if( (x+dx) > ptMaxPos.x )
							ptMaxPos.x = x+dx;
						if( (x-dx) < ptMinPos.x )
							ptMinPos.x = x-dx;
						if( (y+dy) > ptMaxPos.y )
							ptMaxPos.y = y+dy;
						if( (y-dy) < ptMinPos.y )
							ptMinPos.y = y-dy;
					}
					s.m_sel_xi = ptMinPos.x - 8*NM_PER_MIL;
					s.m_sel_yi = ptMinPos.y - 8*NM_PER_MIL;
					s.m_sel_xf = ptMaxPos.x + 8*NM_PER_MIL;
					s.m_sel_yf = ptMaxPos.y + 8*NM_PER_MIL;
					s.WriteFootprint( lib_file );

					s.m_source.Empty();
					s.m_desc.Empty();
					if(s.m_tl)
					{
						s.m_tl->RemoveAllTexts();
					}
				}
				else
				{
					CString str;
					str.Format(_T("    missing pins, aborting conversion\r\n"));
					edit_ctrl->ReplaceSel( str );
				}
			}
			else if( state == MODULE && keystr == _T("NAME"))
			{
				CString name = p[0];
				s.m_name = name;
				CString str;
				str.Format(_T("  converting part \"%s\"\r\n"), name );
				edit_ctrl->ReplaceSel( str );
			}
			else if( state == MODULE && keystr == _T("OUTLINE"))
			{
				if( p[0] == _T("BOX"))
				{
					int xi = my_atof(p[1]) * mult;
					int yi = -(my_atof(p[2]) * mult);
					int xf = my_atof(p[3]) * mult;
					int yf = -(my_atof(p[4]) * mult);
					int np = s.m_outline_poly.GetSize();
					s.m_outline_poly.SetSize(np+1);
					s.m_outline_poly[np].Start( 0, 7*NM_PER_MIL, 0, xi, yi, 0, NULL, NULL );
					s.m_outline_poly[np].AppendCorner( xf, yi );
					s.m_outline_poly[np].AppendCorner( xf, yf );
					s.m_outline_poly[np].AppendCorner( xi, yf );
					s.m_outline_poly[np].Close();
					if( xi > ptMaxPos.x )
						ptMaxPos.x = xi;
					if( xi < ptMinPos.x )
						ptMinPos.x = xi;
					if( yi > ptMaxPos.y )
						ptMaxPos.y = yi;
					if( yi < ptMinPos.y )
						ptMinPos.y = yi;
					if( xf > ptMaxPos.x )
						ptMaxPos.x = xf;
					if( xf < ptMinPos.x )
						ptMinPos.x = xf;
					if( yf > ptMaxPos.y )
						ptMaxPos.y = yf;
					if( yf < ptMinPos.y )
						ptMinPos.y = yf;
				}
				else if( p[0] == _T("LINE"))
				{
					int xi = my_atof(p[1]) * mult;
					int yi = -(my_atof(p[2]) * mult);
					int xf = my_atof(p[3]) * mult;
					int yf = -(my_atof(p[4]) * mult);
					int np = s.m_outline_poly.GetSize();

					s.m_outline_poly.SetSize(np+1);
					s.m_outline_poly[np].Start( 0, 7*NM_PER_MIL, 0, xi, yi, 0, NULL, NULL );
					s.m_outline_poly[np].AppendCorner( xf, yf );
					if( xi > ptMaxPos.x )
						ptMaxPos.x = xi;
					if( xi < ptMinPos.x )
						ptMinPos.x = xi;
					if( yi > ptMaxPos.y )
						ptMaxPos.y = yi;
					if( yi < ptMinPos.y )
						ptMinPos.y = yi;
					if( xf > ptMaxPos.x )
						ptMaxPos.x = xf;
					if( xf < ptMinPos.x )
						ptMinPos.x = xf;
					if( yf > ptMaxPos.y )
						ptMaxPos.y = yf;
					if( yf < ptMinPos.y )
						ptMinPos.y = yf;
				}
				else if( p[0] == _T("ARC"))
				{
					// make an arc with a polyline
					int xc = my_atof(p[1]) * mult;
					int yc = -(my_atof(p[2]) * mult);
					int r = my_atof(p[3]) * mult;
					int quadrant = my_atoi(p[4]);
					int np = s.m_outline_poly.GetSize();

					s.m_outline_poly.SetSize(np+1);
//-----------------------------------------------
// ****** Bug fixed by Dairyo Gokan ******
					switch(quadrant)
					{
						case 1:	s.m_outline_poly[np].Start( 0, 7*NM_PER_MIL, 0, xc  , yc+r, 0, NULL, NULL );
								s.m_outline_poly[np].AppendCorner( xc+r, yc, CPolyLine::ARC_CW );
								break;
						case 2:	s.m_outline_poly[np].Start( 0, 7*NM_PER_MIL, 0, xc+r, yc  , 0, NULL, NULL );
								s.m_outline_poly[np].AppendCorner( xc  , yc-r, CPolyLine::ARC_CW );
								break;
						case 3:	s.m_outline_poly[np].Start( 0, 7*NM_PER_MIL, 0, xc  , yc-r, 0, NULL, NULL );
								s.m_outline_poly[np].AppendCorner( xc-r, yc, CPolyLine::ARC_CW );
								break;
						case 4:	s.m_outline_poly[np].Start( 0, 7*NM_PER_MIL, 0, xc-r, yc  , 0, NULL, NULL );
								s.m_outline_poly[np].AppendCorner( xc, yc+r, CPolyLine::ARC_CW);
								break;
						case 5:	s.m_outline_poly[np].Start( 0, 7*NM_PER_MIL, 0, xc+r, yc  , 0, NULL, NULL );
								s.m_outline_poly[np].AppendCorner( xc, yc+r, CPolyLine::ARC_CCW );
								break;
						case 6:	s.m_outline_poly[np].Start( 0, 7*NM_PER_MIL, 0, xc  , yc-r, 0, NULL, NULL );
								s.m_outline_poly[np].AppendCorner( xc+r, yc  , CPolyLine::ARC_CCW );
								break;
						case 7:	s.m_outline_poly[np].Start( 0, 7*NM_PER_MIL, 0, xc-r, yc  , 0, NULL, NULL );
								s.m_outline_poly[np].AppendCorner( xc  , yc-r, CPolyLine::ARC_CCW );
								break;
						case 8:	s.m_outline_poly[np].Start( 0, 7*NM_PER_MIL, 0, xc  , yc+r, 0, NULL, NULL );
								s.m_outline_poly[np].AppendCorner( xc-r, yc  , CPolyLine::ARC_CCW);
								break;
						default:ASSERT(FALSE);
								break;
					}
// ****** Bug fixed by Dairyo Gokan ******
//-----------------------------------------------
					if( xc+r > ptMaxPos.x )
						ptMaxPos.x = xc+r;
					if( xc-r < ptMinPos.x )
						ptMinPos.x = xc-r;
					if( yc+r > ptMaxPos.y )
						ptMaxPos.y = yc+r;
					if( yc-r < ptMinPos.y )
						ptMinPos.y = yc-r;
				}
				else if( p[0] == _T("CIRCLE"))
				{
					// make a circle with a polyline
					int xc = my_atof(p[1]) * mult;
					int yc = -(my_atof(p[2]) * mult);
					int r = my_atof(p[3]) * mult;
					int np = s.m_outline_poly.GetSize();
					s.m_outline_poly.SetSize(np+1);
					s.m_outline_poly[np].Start( 0, 7*NM_PER_MIL, 0, xc+r, yc, 0, NULL, NULL );
					s.m_outline_poly[np].AppendCorner( xc, yc+r, CPolyLine::ARC_CCW );
					s.m_outline_poly[np].AppendCorner( xc-r, yc, CPolyLine::ARC_CCW );
					s.m_outline_poly[np].AppendCorner( xc, yc-r, CPolyLine::ARC_CCW );
					s.m_outline_poly[np].Close( CPolyLine::ARC_CCW );
					if( xc+r > ptMaxPos.x )
						ptMaxPos.x = xc+r;
					if( xc-r < ptMinPos.x )
						ptMinPos.x = xc-r;
					if( yc+r > ptMaxPos.y )
						ptMaxPos.y = yc+r;
					if( yc-r < ptMinPos.y )
						ptMinPos.y = yc-r;
				}
			}
			else if( state == MODULE && keystr == _T("MOUNTING"))
			{
				// mounting hole
				int ih = m_mtg_hole.GetSize();
				m_mtg_hole.SetSize( ih+1 );
				if( p[0] == _T("EMPTY"))
					m_mtg_hole[ih].pad_shape = PAD_NONE;
				else if( p[0] == _T("SQUARE"))
					m_mtg_hole[ih].pad_shape = PAD_SQUARE;
				else if( p[0] == _T("ROUND"))
					m_mtg_hole[ih].pad_shape = PAD_ROUND;
				else
					ASSERT(FALSE);
				int x =   my_atof(p[1])*mult; 
				int y = -(my_atof(p[2])*mult);					// ****** Bug fixed by Dairyo Gokan ******
				int diam = my_atof(p[3])*mult; 
				int pad_diam = my_atof(p[4])*mult;				// only used for ROUND or SQUARE
				m_mtg_hole[ih].x = x;
				m_mtg_hole[ih].y = y;
				m_mtg_hole[ih].diam = diam;
				m_mtg_hole[ih].pad_diam = pad_diam;
				int d = diam;
				if( pad_diam > diam )
					d = pad_diam;
				int r = d/2;
				if( x+r > ptMaxPos.x )
					ptMaxPos.x = x+r;
				if( x-r < ptMinPos.x )
					ptMinPos.x = x-r;
				if( y+r > ptMaxPos.y )
					ptMaxPos.y = y+r;
				if( y-r < ptMinPos.y )
					ptMinPos.y = y-r;
			}
			else if( state == MODULE && keystr == _T("TEXT"))
			{
				int		x,y,w,h,t,angle;

				x     =  my_atof(p[0])*mult;
				y     =-(my_atof(p[1])*mult);
				angle =  my_atof(p[2]);
				h     =  my_atof(p[4])*mult;
				t     =  my_atof(p[6])*mult;
				w     =  h*p[7].GetLength();

				if(t==0)
					t=7 * NM_PER_MIL;

				switch(angle)
				{
					case 0:		x-=w/2,	y-=h/2;
								break;
					case 90:	x-=h/2,	y+=w/2;
								break;
					case 180:	x+=w/2,	y+=h/2;
								break;
					case 270:	x+=h/2,	y-=w/2;
								break;
					default:	ASSERT(FALSE);
								break;
				}
				if(p[7] == _T("REF"))
				{
					s.m_ref_size=h;									/* iԍ(REF)̍̕		*/
					s.m_ref_xi  =x;									/* iԍ(REF)̑΃ItZbg(x)	*/
					s.m_ref_yi  =y;									/* iԍ(REF)̑΃ItZbg(y)	*/
					s.m_ref_angle= angle;							/* iԍ(REF)̉]px(CW)		*/
					s.m_ref_w   =t;									/* iԍ(REF)̕`g[X	*/
				}
				else if(p[7] == _T("VAL"))
				{
					s.m_value_size =h;								/* i^(VAL)̍̕		*/
					s.m_value_xi   =x;								/* i^(VAL)̑΃ItZbg(x) */
					s.m_value_yi   =y;								/* i^(VAL)̑΃ItZbg(y) */
					s.m_value_angle=angle;							/* i^(VAL)̉]px(CW)		*/
					s.m_value_w    =t;								/* i^(VAL)̕`g[X	*/
				}
				else if(!p[7].IsEmpty())
				{
					if(!s.m_tl)
						s.m_tl=new CTextList;
					s.m_tl->AddText(x,y,angle,0,0,LAY_SILK_TOP,h,t,&p[7],FALSE);
				}
			}
			else if( state == MODULE && keystr == _T("PAD") 
				&& ( p[0] == _T("DEF") || p[0] == _T("STEP")) )
			{
				// add pad to shape
//				int pin_num;
				CString pin_str;
				if( p[0] == _T("DEF"))
				{
					current_x = my_atof(p[1]) * mult;
					current_y = my_atof(p[2]) * mult;
					current_angle = my_atof(p[3]);
					CString pad_name = p[4];
					void * ptr;
					BOOL found = mod_padstack_map.Lookup( pad_name, ptr );
					if( !found )
						ASSERT(FALSE);
					current_ps = (mod_padstack *)ptr;
					pin_str = p[5];
				}
				else if( p[0] == _T("STEP"))
				{
					CString step_str = p[1];
					CString step = step_str.Right( step_str.GetLength()-2 );
					int is = my_atof( step ) * mult;
					if( step_str.Left(2) == _T("X+"))
						current_x = current_x + is;
					if( step_str.Left(2) == _T("X-"))
						current_x = current_x - is;
					if( step_str.Left(2) == _T("Y+"))
						current_y = current_y + is;
					if( step_str.Left(2) == _T("Y-"))
						current_y = current_y - is;
					pin_str = p[2];
				}
				if( pin_str != _T("") )
				{
					int pin_num = s.m_padstack.GetSize() + 1;
					bool	bTop,bInner,bBottom;

					s.m_padstack.SetSize(pin_num);
					s.m_padstack[pin_num-1].exists = TRUE;
					s.m_padstack[pin_num-1].name = pin_str;
					s.m_padstack[pin_num-1].hole_size = current_ps->hole_diam;
					s.m_padstack[pin_num-1].x_rel = current_x;
					s.m_padstack[pin_num-1].y_rel = -(current_y);	// since Ivex reverses y-axis
					s.m_padstack[pin_num-1].angle = current_angle;
					s.m_padstack[pin_num-1].top.shape = PAD_NONE;
					s.m_padstack[pin_num-1].inner.shape = PAD_NONE;
					s.m_padstack[pin_num-1].bottom.shape = PAD_NONE;

					bTop=bInner=bBottom=false;
					for(int nRep=0;nRep<current_ps->entry.GetSize();nRep++)
					{
						const pad_stack_entry	&padshape=current_ps->entry[nRep];

						if(padshape.layer_mask & (LAY_T | LAY_C) )
						{
							if(!bTop)
							{
								s.m_padstack[pin_num-1].top.shape  = padshape.pad_shape;
								s.m_padstack[pin_num-1].top.size_l = padshape.x_wid/2;
								s.m_padstack[pin_num-1].top.size_r = padshape.x_wid/2;
								s.m_padstack[pin_num-1].top.size_h = padshape.y_wid;
								bTop=true;
							}
//							else
//								ASSERT(FALSE);					/* Detected double registration */
						}
						if( current_ps->entry[0].layer_mask & LAY_I || current_ps->hole_diam > 0 )
						{
							if(!bInner)
							{
								// inner pads are forced round
								s.m_padstack[pin_num-1].inner.shape  = PAD_ROUND;
								s.m_padstack[pin_num-1].inner.size_l = min(padshape.x_wid,padshape.y_wid)/2;
								s.m_padstack[pin_num-1].inner.size_r = min(padshape.x_wid,padshape.y_wid)/2;
								s.m_padstack[pin_num-1].inner.size_h = min(padshape.x_wid,padshape.y_wid);
								bInner=true;
							}
//							else
//								ASSERT(FALSE);					/* Detected double registration */
						}
						if( current_ps->entry[0].layer_mask & LAY_B )
						{
							if(!bBottom)
							{
								s.m_padstack[pin_num-1].bottom.shape  = padshape.pad_shape;
								s.m_padstack[pin_num-1].bottom.size_l = padshape.x_wid/2;
								s.m_padstack[pin_num-1].bottom.size_r = padshape.x_wid/2;
								s.m_padstack[pin_num-1].bottom.size_h = padshape.y_wid;
								bBottom=true;
							}
//							else
//								ASSERT(FALSE);					/* Detected double registration */
						}
					}
				}
				else
				{
					// illegal pin number
					CString str;
					str.Format(_T("    illegal pin \"%s\", abort conversion\r\n"), pin_str );
					edit_ctrl->ReplaceSel( str );
					state = IDLE;
				}
			}
		}
		ret = mod_file->ReadString( instr );
	}
	// now delete map of padstacks
	POSITION pos;
	CString key;
	void * ptr;
	for( pos = mod_padstack_map.GetStartPosition(); pos != NULL; )
	{
		mod_padstack_map.GetNextAssoc( pos, key, ptr );
		delete ptr;
		return 0;
	}
	return 0;
}


